#include <windows.h>
#include "pmdwinx.h"


static HINSTANCE hPLUGIN = NULL;
static HINSTANCE hPMDWIN = NULL;
static HINSTANCE hWINFMP = NULL;
static enum {
	OFF_MODE,
	FMP_MODE,
	PMD_MODE
} fmpmdmode;

static initFMP = 0;
static initPMD = 0;
static void TermPMD(void)
{
	if (hPMDWIN)
	{
		FreeLibrary(hPMDWIN);
		hPMDWIN = 0;
		initPMD = 0;
	}
}

static void TermFMP(void)
{
	if (hWINFMP)
	{
		FreeLibrary(hWINFMP);
		hWINFMP = 0;
		initFMP = 0;
	}
}


static bool (WINAPI *pmdwininitp)(char *path);
static bool (WINAPI *fmp_initp)(char *path);
bool WINAPI pmdwininitx(char *path)
{
	bool ret = 0;
	if (!initPMD && hPMDWIN)
	{	
		initPMD = pmdwininitp(path);
		if (!initPMD) TermPMD();
	}
	if (!initFMP && hWINFMP)
	{
		initFMP = fmp_initp(path);
		if (!initFMP) TermFMP();
	}
	return initFMP || initPMD;
}

static bool (WINAPI *setpcmdirp)(char **path);
static bool (WINAPI *fmp_setpcmdirp)(char **path);
bool WINAPI setpcmdirx(char **path)
{
	if (hPMDWIN) setpcmdirp(path);
	if (hWINFMP) fmp_setpcmdirp(path);
	return 1;
}

void WINAPI pmdwintermx(void)
{
	fmpmdmode = OFF_MODE;
}

static int (WINAPI *music_loadp)(char *filename);
static int (WINAPI *fmp_loadp)(char *filename);
int WINAPI music_loadx(char *filename)
{
	int ret = hPMDWIN ? music_loadp(filename) : ERR_OTHER ;
	switch (ret)
	{
		case PMDWIN_OK:
		case WERNING_PPC_ALREADY_LOAD:
		case WERNING_P86_ALREADY_LOAD:
		case WERNING_PPS_ALREADY_LOAD:
		case WERNING_PPZ1_ALREADY_LOAD:
		case WERNING_PPZ2_ALREADY_LOAD:
			fmpmdmode = PMD_MODE;
			return ret;
		default:
		case ERR_OTHER:
			/* ? ? ? */
		case ERR_OPEN_MUSIC_FILE:
		case ERR_WRONG_MUSIC_FILE:
		case ERR_OPEN_PPC_FILE:
		case ERR_WRONG_PPC_FILE:
		case ERR_OPEN_P86_FILE:
		case ERR_WRONG_P86_FILE:
		case ERR_OPEN_PPS_FILE:
		case ERR_WRONG_PPS_FILE:
		case ERR_OPEN_PPZ1_FILE:
		case ERR_WRONG_PPZ1_FILE:
		case ERR_OPEN_PPZ2_FILE:
		case ERR_WRONG_PPZ2_FILE:
		case ERR_OUT_OF_MEMORY:
			if (!hWINFMP) return ret;
			fmpmdmode = FMP_MODE;
			return fmp_loadp(filename);
	}
}

static bool (WINAPI *getlengthp)(char *p1, int *p2, int *p3);
static bool (WINAPI *fmp_getlengthp)(char *p1, int *p2, int *p3);
bool WINAPI getlengthx(char *p1, int *p2, int *p3) {
	int ln, lp;
	if (!initPMD || !initFMP) pmdwininitx(0);
	if (hPMDWIN && getlengthp(p1, &ln, &lp))
	{	
		if (p2) *p2 = ln;
		if (p3) *p3 = lp;
		return 1;
	}
	if (hWINFMP && fmp_getlengthp(p1, &ln, &lp))
	{	
		if (p2) *p2 = ln;
		if (p3) *p3 = lp;
		return 1;
	}
	if (p2) *p2 = 0;
	if (p3) *p3 = 0;
	return 0;
}


static void (WINAPI *setppsusep)(bool value);
void WINAPI setppsusex(bool value)
{
	switch (fmpmdmode) {
		case PMD_MODE:
			setppsusep(value);
	}
}
static void (WINAPI *setpmd86pcmmodep)(bool value);
void WINAPI setpmd86pcmmodex(bool value)
{
	switch (fmpmdmode) {
		case PMD_MODE:
			setpmd86pcmmodep(value);
	}
}
static void (WINAPI *setrhythmwithssgeffectp)(bool value);
void WINAPI setrhythmwithssgeffectx(bool value)
{
	switch (fmpmdmode) {
		case PMD_MODE:
			setrhythmwithssgeffectp(value);
	}
}

#define P0X(f) \
static void (WINAPI *music_##f##p)(void); \
static void (WINAPI *fmp_##f##p)(void); \
void WINAPI music_##f##x(void) { \
	switch (fmpmdmode) { \
		case PMD_MODE: \
			music_##f##p(); \
			break; \
		case FMP_MODE: \
			fmp_##f##p(); \
			break; \
	} \
}

#define P0R(f, rt) \
static rt (WINAPI *##f##p)(void); \
static rt (WINAPI *fmp_##f##p)(void); \
rt WINAPI f##x(void) { \
	switch (fmpmdmode) { \
		case PMD_MODE: \
			return f##p(); \
		case FMP_MODE: \
			return fmp_##f##p(); \
	} \
	return 0; \
}

#define P1(f, pt) \
static void (WINAPI *##f##p)(pt p1); \
static void (WINAPI *fmp_##f##p)(pt p1); \
void WINAPI f##x(pt p1) { \
	switch (fmpmdmode) { \
		case PMD_MODE: \
			f##p(p1); \
			break; \
		case FMP_MODE: \
			fmp_##f##p(p1); \
			break; \
	} \
}

#define P1R(f, rt, pt) \
static rt (WINAPI *##f##p)(pt p1); \
static rt (WINAPI *fmp_##f##p)(pt p1); \
rt WINAPI f##x(pt p1) { \
	switch (fmpmdmode) { \
		case PMD_MODE: \
			return  f##p(p1); \
		case FMP_MODE: \
			return  fmp_##f##p(p1); \
	} \
	return 0; \
}

#define P2(f, pt, pt2) \
static void (WINAPI *##f##p)(pt p1, pt2 p2); \
static void (WINAPI *fmp_##f##p)(pt p1, pt2 p2); \
void WINAPI f##x(pt p1, pt2 p2) { \
	switch (fmpmdmode) { \
		case PMD_MODE: \
			f##p(p1, p2); \
			break; \
		case FMP_MODE: \
			fmp_##f##p(p1, p2); \
			break; \
	} \
}

#define P3R(f, rt, pt, pt2, pt3) \
static rt (WINAPI *##f##p)(pt p1, pt2 p2, pt3 p3); \
static rt (WINAPI *fmp_##f##p)(pt p1, pt2 p2, pt3 p3); \
rt WINAPI f##x(pt p1, pt2 p2, pt3 p3) { \
	switch (fmpmdmode) { \
		case PMD_MODE: \
			return f##p(p1, p2, p3); \
		case FMP_MODE: \
			return fmp_##f##p(p1, p2, p3); \
	} \
	return 0; \
}

P1R(loadrhythmsample, bool, char *)

P1(setfmvoldown, int)
P1(setssgvoldown, int)
P1(setrhythmvoldown, int)
P1(setadpcmvoldown, int)
P1(setppzvoldown, int)

P1(setpcmrate, int)
P0R(getloopcount, int)

P0X(start)
P0X(stop)
P1(setpos, int)
P0R(getpos, int)

P2(getpcmdata, short *, int)

P1(fadeout2, int)


static int (WINAPI *fgetmemo3p)(char *dest, char *filename, int al);
static int (WINAPI *fmp_fgetcomment3p)(char **dest, char *filename);
int WINAPI fgettitlex(char *dest, char *filename) {
	if (!initPMD || !initFMP) pmdwininitx(0);
	if (hWINFMP) {
		char line1[80+2];        // Ps
		char line2[80+2];        // Qs
		char line3[80+2];        // Rs
		char *adest[3];
		adest[0] = line1;
		adest[1] = line2;
		adest[2] = line3;
		if (fmp_fgetcomment3p(adest, filename) == 0) {
			if (line1[0]) {	/* G[AĂȂ */
				lstrcpy(dest, line1);
				return PMDWIN_OK;
			}
		}
	}
	if (hPMDWIN && fgetmemo3p(dest, filename, 1) == PMDWIN_OK) {
		return PMDWIN_OK;
	}
	return ERR_OPEN_MUSIC_FILE;
}

static int (WINAPI *fmp_maskonp)(bool rhythm_flag, int ah);
static int (WINAPI *maskonp)(int ch);
int WINAPI maskonx(int ch)
{
	switch (fmpmdmode) {
		case PMD_MODE:
			return maskonp(ch);
		case FMP_MODE:
			switch (ch) {
				/* FM */
				case 0: case 1: case 2:
				case 3: case 4: case 5:
					return fmp_maskonp(0, ch);
				/* SSG */
				case 6: case 7: case 8:
					return fmp_maskonp(0, ch - 6 + 18);
				/* ADPCM */
				case 9:
					return fmp_maskonp(0, 21);
				/* RHTYM */
				case 10:
					return fmp_maskonp(1, 63);
				/* FMCH3EX */
				case 11: case 12: case 13:
					return fmp_maskonp(0, ch - 11 + 6);
				/* PPZ8 */
				case 16: case 17: case 18: case 19:
				case 20: case 21: case 22: case 23:
					return fmp_maskonp(0, ch - 16 + 10);
			}
	}
	return 0;
}

static int (WINAPI *fmp_maskoffp)(bool rhythm_flag, int ch);
static int (WINAPI *maskoffp)(int ch);
int WINAPI maskoffx(int ch)
{
	switch (fmpmdmode) {
		case PMD_MODE:
			return maskoffp(ch);
		case FMP_MODE:
			switch (ch) {
				/* FM */
				case 0: case 1: case 2:
				case 3: case 4: case 5:
					return fmp_maskoffp(0, ch);
				/* SSG */
				case 6: case 7: case 8:
					return fmp_maskoffp(0, ch - 6 + 18);
				/* ADPCM */
				case 9:
					return fmp_maskoffp(0, 21);
				/* RHTYM */
				case 10:
					return fmp_maskoffp(1, 63);
				/* FMCH3EX */
				case 11: case 12: case 13:
					return fmp_maskoffp(0, ch - 11 + 6);
				/* PPZ8 */
				case 16: case 17: case 18: case 19:
				case 20: case 21: case 22: case 23:
					return fmp_maskoffp(0, ch - 16 + 10);
			}
	}
	return 0;
}

bool WINAPI isplayingx(void)
{
	return (getloopcountx() != -1);
}

static void TermFMPMD(void)
{
	fmpmdmode = OFF_MODE;
	TermPMD();
	TermFMP();
}


HINSTANCE SearchDll(char *name) {
	HINSTANCE ret;
	char path1[MAX_PATH], path2[MAX_PATH];
	LPTSTR fnoff;
	GetModuleFileName(hPLUGIN, path1, MAX_PATH);
	GetFullPathName(path1, MAX_PATH, path2, &fnoff);
	path2[(fnoff - path2)] = '\0';
	lstrcat(path2, name);
	ret = LoadLibrary(path2);
	if (ret) return ret;
	ret = LoadLibrary(name);
	return ret;
}

#define RESOLVE(f) \
	{ \
		f##p = GetProcAddress(hPMDWIN, #f); \
		if (!##f##p) break; \
	}

static void ResolvePMD(void) {
	hPMDWIN = SearchDll("PMDWin.dll");
	if (!hPMDWIN) return;
	while (1) {
		RESOLVE(pmdwininit);
		RESOLVE(music_load);
		RESOLVE(music_start);
		RESOLVE(music_stop);
		RESOLVE(setppsuse);
		RESOLVE(setpmd86pcmmode);
		RESOLVE(setrhythmwithssgeffect);
		RESOLVE(fgetmemo3);

		RESOLVE(loadrhythmsample);
		RESOLVE(getloopcount);
		RESOLVE(getlength);
		RESOLVE(maskon);
		RESOLVE(maskoff);

		RESOLVE(setpcmdir);
		RESOLVE(setfmvoldown);
		RESOLVE(setssgvoldown);
		RESOLVE(setrhythmvoldown);
		RESOLVE(setadpcmvoldown);
		RESOLVE(setppzvoldown);
		RESOLVE(setpcmrate);
		RESOLVE(getpcmdata);
		RESOLVE(fadeout2);
		RESOLVE(setpos);
		RESOLVE(getpos);

		return;
	}
	FreeLibrary(hPMDWIN);
	hPMDWIN = 0;
}

#undef RESOLVE
#define RESOLVE(f) \
	{ \
		fmp_##f##p = GetProcAddress(hWINFMP, "fmp_" #f); \
		if (!fmp_##f##p) break; \
	}


static void ResolveFMP(void) {
	hWINFMP = SearchDll("WinFMP.dll");
	if (!hWINFMP) return;
	while (1) {
		RESOLVE(init);
		RESOLVE(start);
		RESOLVE(stop);
		RESOLVE(load);
		RESOLVE(fgetcomment3);

		RESOLVE(getloopcount);
		RESOLVE(getlength);
		RESOLVE(maskon);
		RESOLVE(maskoff);

		RESOLVE(loadrhythmsample);
		RESOLVE(setpcmdir);
		RESOLVE(setfmvoldown);
		RESOLVE(setssgvoldown);
		RESOLVE(setrhythmvoldown);
		RESOLVE(setadpcmvoldown);
		RESOLVE(setppzvoldown);
		RESOLVE(setpcmrate);
		RESOLVE(getpcmdata);
		RESOLVE(fadeout2);
		RESOLVE(setpos);
		RESOLVE(getpos);
		return;
	}
	FreeLibrary(hWINFMP);
	hWINFMP = 0;
}

char *InitializeFMPMD(void)
{
	ResolvePMD();
	ResolveFMP();
	if (hPMDWIN && hWINFMP)
		return "M;M2;MZ;MP;MS\0PMD music files (*.m;*m2;*.mz;*.mp;*.ms)\0OPI;OVI;OZI\0FMP music files (*.opi;*ovi;*.ozi)\0\0";
	if (hPMDWIN)
		return "M;M2;MZ;MP;MS\0PMD music files (*.m;*m2;*.mz;*.mp;*.ms)\0\0";
	if (hWINFMP)
		return "OPI;OVI;OZI\0FMP music files (*.opi;*ovi;*.ozi)\0\0";
	return 0;
}

BOOL WINAPI DllMain(
	HINSTANCE hinstDLL,
	DWORD fdwReason,
	LPVOID lpvReserved
)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		hPLUGIN = hinstDLL;
	}
	else if (fdwReason == DLL_PROCESS_DETACH)
	{
		TermFMPMD();
	}
	return TRUE;
}
